# -*- coding: utf-8 -*-
"""Contains classes for handling atmospheric temperature lapse rates."""
import abc
import numbers

import numpy as np
from scipy.interpolate import interp1d
from typhon.physics import vmr2mixing_ratio

from konrad import constants
from konrad.component import Component
from konrad.physics import saturation_pressure


class LapseRate(Component, metaclass=abc.ABCMeta):
    """Base class for all lapse rate handlers."""
    @abc.abstractmethod
    def __call__(self, atmosphere):
        """Return the atmospheric lapse rate.

        Parameters:
              T (ndarray): Atmospheric temperature [K].
              p (ndarray): Atmospheric pressure [Pa].

        Returns:
              ndarray: Temperature lapse rate [K/m].
        """


class MoistLapseRate(LapseRate):
    """Moist adiabatic temperature lapse rate."""
    def __init__(self, fixed=False):
        """Initialize a moist-adiabatic lapse rate component.

        Parameters:
            fixed (bool): If `True` the moist adiabatic lapse rate is only
                calculated for the first time step and kept constant
                afterwards.
        """
        self.fixed = fixed
        self._lapse_cache = None

    def __call__(self, atmosphere):
        if self._lapse_cache is not None:
            return self._lapse_cache

        T = atmosphere['T'][0, :]
        p = atmosphere['plev'][:]
        phlev = atmosphere['phlev'][:]

        # Use short formula symbols for physical constants.
        g = constants.earth_standard_gravity
        L = constants.heat_of_vaporization
        Rd = constants.specific_gas_constant_dry_air
        Rv = constants.specific_gas_constant_water_vapor
        Cp = constants.isobaric_mass_heat_capacity

        gamma_d = g / Cp  # dry lapse rate

        w_saturated = vmr2mixing_ratio(saturation_pressure(T) / p)

        gamma_m = (gamma_d * ((1 + (L * w_saturated) / (Rd * T)) /
                              (1 + (L**2 * w_saturated) / (Cp * Rv * T**2))
                              )
        )
        lapse = interp1d(p, gamma_m, fill_value='extrapolate')(phlev[:-1])

        if self.fixed:
            self._lapse_cache = lapse

        return lapse


class FixedLapseRate(LapseRate):
    """Fixed linear lapse rate through the whole atmosphere."""
    def __init__(self, lapserate=0.0065):
        """Create a handler with fixed linear temperature lapse rate.

        Parameters:
              lapserate (float or ndarray): Critical lapse rate [K/m].
        """
        self.lapserate = lapserate

    def __call__(self, atmosphere):
        if isinstance(self.lapserate, numbers.Number):
            T = atmosphere['T'][0, :]
            return self.lapserate * np.ones(T.size)
        elif isinstance(self.lapserate, np.ndarray):
            return self.lapserate



# -*- coding: utf-8 -*-
"""This module contains classes handling different treatments of ozone."""

import os
import abc
import logging
import numpy as np
from netCDF4 import Dataset
from scipy.interpolate import interp1d
from konrad.component import Component

__all__ = [
    'Ozone',
    'OzonePressure',
    'OzoneHeight',
    'OzoneNormedPressure',
    'OzoneCariolle',
    'OzoneSiRaChA',
]

logger = logging.getLogger(__name__)


class Ozone(Component, metaclass=abc.ABCMeta):
    """Base class to define abstract methods for ozone treatments."""

    def __init__(self):
        """
        Parameters:
            initial_ozone (ndarray): initial ozone vmr profile
        """
        self['initial_ozone'] = (('plev',), None)

    @abc.abstractmethod
    def __call__(self, atmosphere, convection, timestep, zenith):
        """Updates the ozone profile within the atmosphere class.

        Parameters:
            atmosphere (konrad.atmosphere): atmosphere model containing ozone
                concentration profile, height, temperature, pressure and half
                pressure levels at the current timestep
            convection (konrad.convection): convection scheme
            timestep (float): timestep of run [days]
            zenith (float): solar zenith angle,
                angle of the Sun to the vertical [degrees]
        """


class OzonePressure(Ozone):
    """Ozone fixed with pressure, no adjustment needed."""
    def __call__(self, **kwargs):
        return


class OzoneHeight(Ozone):
    """Ozone fixed with height."""
    def __init__(self):
        self._f = None

    def __call__(self, atmosphere, **kwargs):
        if self._f is None:
            self._f = interp1d(
                atmosphere['z'][0, :],
                atmosphere['O3'],
                fill_value='extrapolate',
            )
        atmosphere['O3'] = (('time', 'plev'), self._f(atmosphere['z'][0, :]))


class OzoneNormedPressure(Ozone):
    """Ozone shifts with the normalisation level (chosen to be the convective
    top)."""
    def __init__(self, norm_level=None):
        """
        Parameters:
            norm_level (float): pressure for the normalisation
                normally chosen as the convective top pressure at the start of
                the simulation [Pa]
        """
        self.norm_level = norm_level
        self._f = None

    def __call__(self, atmosphere, convection, **kwargs):
        if self.norm_level is None:
            self.norm_level = convection.get('convective_top_plev')[0]
            # TODO: what if there is no convective top

        if self._f is None:
            self._f = interp1d(
                atmosphere['plev'] / self.norm_level,
                atmosphere['O3'][0, :],
                fill_value='extrapolate',
            )

        norm_new = convection.get('convective_top_plev')[0]

        atmosphere['O3'] = (
            ('time', 'plev'),
            self._f(atmosphere['plev'] / norm_new).reshape(1, -1)
        )


class OzoneCariolle(Ozone):
    """Implementation of the Cariolle ozone scheme for the tropics.
    """
    def __init__(self, w=0):
        """
        Parameters:
            w (ndarray / int / float): upwelling velocity [mm / s]
        """
        super().__init__()
        self.w = w * 86.4  # in m / day

    def ozone_transport(self, o3, z, convection):
        """Rate of change of ozone is calculated based on the ozone gradient
        and an upwelling velocity.

        Parameters:
            o3 (ndarray): ozone concentration [ppv]
            z (ndarray): height [m]
            convection (konrad.convection): to get the convective top index
        Returns:
            ndarray: change in ozone concentration [ppv / day]
        """
        if self.w == 0:
            return np.zeros(len(z))

        if isinstance(self.w, np.ndarray):
            w_array = self.w
        else:  # w is a single value
            # apply transport only above convective top
            w = self.w
            numlevels = len(z)
            contopi = convection.get('convective_top_index')[0]
            if np.isnan(contopi):
                # No convective top index found; do not apply transport term
                return np.zeros(numlevels)
            contopi = int(np.round(contopi))
            w_factor = np.ones(numlevels)
            w_factor[:contopi] = 0
            w_array = w*w_factor

        do3dz = (o3[1:] - o3[:-1]) / np.diff(z)
        do3dz = np.hstack(([0], do3dz))

        return -w_array * do3dz

    def get_params(self, p):
        cariolle_data = Dataset(
            os.path.join(os.path.dirname(__file__),
                         '../Cariolle_data.nc'))
        p_data = cariolle_data['p'][:]
        alist = []
        for param_num in range(1, 8):
            a = cariolle_data[f'A{param_num}'][:]
            alist.append(interp1d(p_data, a, fill_value='extrapolate')(p))
        return alist

    def __call__(self, atmosphere, convection, timestep, *args, **kwargs):

        from SiRaChA.utils import overhead_molecules

        T = atmosphere['T'][0, :]
        p = atmosphere['plev']  # [Pa]
        phlev = atmosphere['phlev']
        o3 = atmosphere['O3'][0, :]  # moles of ozone / moles of air
        z = atmosphere['z'][0, :]  # m

        o3col = overhead_molecules(o3, p, phlev, z, T
                                   ) * 10 ** -4  # in molecules / cm2

        A1, A2, A3, A4, A5, A6, A7 = self.get_params(p)
        # A7 is in molecules / cm2
        # tendency of ozone volume mixing ratio per second
        do3dt = A1 + A2*(o3 - A3) + A4*(T - A5) + A6*(o3col - A7)

        # transport term
        transport_ox = self.ozone_transport(o3, z, convection)

        atmosphere['O3'] = (
            ('time', 'plev'),
            (o3 + (do3dt * 24 * 60**2 + transport_ox) * timestep).reshape(1, -1)
        )


class OzoneSiRaChA(OzoneCariolle):

    def __init__(self, w=0):
        """
        Parameters:
            w (ndarray / int / float): upwelling velocity [mm / s]
        """
        super().__init__()

        from SiRaChA import SiRaChA

        self.w = w * 86.4  # in m / day
        self._ozone = SiRaChA()

    def __call__(self, atmosphere, convection, timestep, zenith, *args,
                 **kwargs):

        o3 = atmosphere['O3'][-1, :]
        z = atmosphere['z'][-1, :]
        p, phlev = atmosphere['plev'], atmosphere['phlev']
        T = atmosphere['T'][-1, :]
        source, sink_ox, sink_nox, sink_hox = self._ozone.tendencies(
            z, p, phlev, T, o3, zenith)
        transport_ox = self.ozone_transport(o3, z, convection)
        do3dt = source - sink_ox - sink_nox + transport_ox - sink_hox

        atmosphere['O3'] = (
            ('time', 'plev'),
            (o3 + do3dt * timestep).reshape(1, -1)
        )

        for term, tendency in [('ozone_source', source),
                               ('ozone_sink_ox', sink_ox),
                               ('ozone_sink_nox', sink_nox),
                               ('ozone_transport', transport_ox),
                               ('ozone_sink_hox', sink_hox)
                               ]:
            if term in self.data_vars:
                self.set(term, tendency)
            else:
                self.create_variable(term, tendency)



# -*- coding: utf-8 -*-
"""Module containing classes describing different radiation models.
"""
from .radiation import Radiation
from .rrtmg import RRTMG


__all__ = [s for s in dir() if not s.startswith('_')]



"""This module contains classes for handling humidity."""
import logging

from konrad.component import Component
from konrad.utils import prefix_dict_keys
from konrad.physics import (relative_humidity2vmr, vmr2relative_humidity)
from .stratosphere import *
from .relative_humidity import *


logger = logging.getLogger(__name__)


class FixedRH(Component):
    """Preserve the relative humidity profile under temperature changes."""
    def __init__(self, rh_func=None, stratosphere_coupling=None):
        """Create a humidity handler.

        Parameters:
            rh_func (callable): Callable that describes the vertical
                relative humidity distribution.
            stratosphere_coupling (callable):
        """
        if stratosphere_coupling is None:
            self._stratosphere_coupling = ColdPointCoupling()
        else:
            self._stratosphere_coupling = stratosphere_coupling

        self._rh_func = rh_func
        self._rh_profile = None

    @property
    def attrs(self):
        # Overrides ``Component.attrs`` by returning a composite of the
        # attribtues of both ``rh_func`` and ``stratosphere_coupling``.
        # The returned attributes are prefixed with their parent-attribute's
        # name for clarity.
        attrs = dict(
            **prefix_dict_keys(
                getattr(self._rh_func, 'attrs', {}), 'rh_func'
            ),
            **prefix_dict_keys(
                self._stratosphere_coupling.attrs, 'stratosphere_coupling',
            )
        )
        attrs['rh_func/class'] = self.rh_func
        attrs['stratosphere_coupling/class'] = self.stratosphere_coupling

        return attrs

    def hash_attributes(self):
        # Make sure that non-``Component`` attributes do not break hashing.
        return hash(tuple(
            attr.hash_attributes()
            for attr in (self._rh_func, self._stratosphere_coupling)
            if hasattr(attr, 'hash_attributes')
        ))

    @property
    def rh_func(self):
        return type(self._rh_func).__name__

    @property
    def stratosphere_coupling(self):
        return type(self._stratosphere_coupling).__name__

    def adjust_humidity(self, atmosphere, **kwargs):
        """Determine the humidity profile based on atmospheric state.

        Parameters:
            TODO: Write docstring.

        Returns:
            ndarray: Water vapor profile [VMR].
        """
        if self._rh_func is not None:
           rh_profile = self._rh_func(atmosphere, **kwargs)
        else:
            if self._rh_profile is None:
                self._rh_profile = vmr2relative_humidity(
                        vmr=atmosphere['H2O'][-1],
                        pressure=atmosphere['plev'],
                        temperature=atmosphere['T'][-1]
                    )
            rh_profile = self._rh_profile

        atmosphere['H2O'][-1, :] = relative_humidity2vmr(
            relative_humidity=rh_profile,
            pressure=atmosphere['plev'],
            temperature=atmosphere['T'][-1]

        )
        self._stratosphere_coupling.adjust_stratospheric_vmr(atmosphere)


class FixedVMR(Component):
    """Keep the water vapor volume mixing ratio constant."""
    def __init__(self, *args, **kwargs):
        if len(args) + len(kwargs) > 0:
            # Allow arguments to be passed for consistent interface but
            # warn the user.
            logger.warning(f'All input arguments to {self} are ignored.')

        # Set both attributes for consistent user interface and netCDF output.
        self.rh_func = 'FixedVMR'
        self.stratosphere_coupling = 'FixedVMR'

    def adjust_humidity(self, atmosphere, **kwargs):
        return



"""This module contains functions related to physics."""
from numbers import Number

import numpy as np
import typhon.physics as typ

from konrad import constants


# TODO: Replace with typhon version after next release (>0.7.0)
def saturation_pressure(temperature):
    r"""Return equilibrium pressure of water with respect to the mixed-phase.

    The equilibrium pressure over water is taken for temperatures above the
    triple point :math:`T_t` the value over ice is taken for temperatures
    below :math:`T_t–23\,\mathrm{K}`.  For intermediate temperatures the
    equilibrium pressure is computed as a combination
    of the values over water and ice according to the IFS documentation:

    .. math::
        e_\mathrm{s} = \begin{cases}
            T > T_t, & e_\mathrm{liq} \\
            T < T_t - 23\,\mathrm{K}, & e_\mathrm{ice} \\
            else, & e_\mathrm{ice}
                + (e_\mathrm{liq} - e_\mathrm{ice})
                \cdot \left(\frac{T - T_t - 23}{23}\right)^2
        \end{cases}

    References:
        IFS Documentation – Cy45r1,
        Operational implementation 5 June 2018,
        Part IV: Physical Processes, Chapter 12, Eq. 12.13,
        https://www.ecmwf.int/node/18714

    Parameters:
        temperature (float or ndarray): Temperature [K].

    See also:
        :func:`~typhon.physics.e_eq_ice_mk`
            Equilibrium pressure of water over ice.
        :func:`~typhon.physics.e_eq_water_mk`
            Equilibrium pressure of water over liquid water.

    Returns:
        float or ndarray: Equilibrium pressure [Pa].
    """
    # Keep track of input type to match the return type.
    is_float_input = isinstance(temperature, Number)
    if is_float_input:
        # Convert float input to ndarray to allow indexing.
        temperature = np.asarray([temperature])

    e_eq_water = typ.e_eq_water_mk(temperature)
    e_eq_ice = typ.e_eq_ice_mk(temperature)

    is_water = temperature > constants.triple_point_water

    is_ice = temperature < (constants.triple_point_water - 23.)

    e_eq = (e_eq_ice + (e_eq_water - e_eq_ice)
            * ((temperature - constants.triple_point_water + 23) / 23)**2
            )
    e_eq[is_ice] = e_eq_ice[is_ice]
    e_eq[is_water] = e_eq_water[is_water]

    return float(e_eq) if is_float_input else e_eq


def relative_humidity2vmr(relative_humidity, pressure, temperature):
    r"""Convert relative humidity into water vapor VMR.

    .. math::
        VMR = \frac{RH \cdot e_s(T)}{p}

    Parameters:
        relative_humidity (float or ndarray): Relative humidity.
        pressure (float or ndarray): Pressue [Pa].
        temperature(float or ndarray): Temperature [K].

    Returns:
        float or ndarray: Water vapor volume mixing ratio [dimensionless].
    """
    return typ.relative_humidity2vmr(
        RH=relative_humidity,
        p=pressure,
        T=temperature,
        e_eq=saturation_pressure,
    )


def vmr2relative_humidity(vmr, pressure, temperature):
    r"""Convert water vapor VMR into relative humidity.

    .. math::
        RH = \frac{VMR \cdot p}{e_s(T)}

    Parameters:
        vmr (float or ndarray): Water vapor volume mixing ratio.
        pressure (float or ndarray): Pressure [Pa].
        temperature (float or ndarray): Temperature [K].

    Returns:
        float or ndarray: Relative humidity [dimensionless].
    """
    return typ.vmr2relative_humidity(
        vmr=vmr,
        p=pressure,
        T=temperature,
        e_eq=saturation_pressure,
    )



"""This module contains classes for handling clouds."""
import abc
import logging

import numpy as np
from scipy.interpolate import interp1d
from sympl import DataArray

logger = logging.getLogger(__name__)


__all__ = [
    'get_rectangular_profile',
    'Cloud',
    'ClearSky',
    'DirectInputCloud',
    'PhysicalCloud',
    'HighCloud',
    'MidLevelCloud',
    'LowCloud',
    'CloudEnsemble',
]


# TODO: Make this a staticmethod of the ``Cloud`` class?
def get_p_data_array(values, units='kg m^-2', numlevels=200):
    """Return a DataArray of values."""
    if isinstance(values, DataArray):
        return values

    elif isinstance(values, np.ndarray):
        if values.shape == (numlevels,):
            return DataArray(values, dims=('mid_levels',),
                             attrs={'units': units})
        else:
            raise ValueError(
                'Cloud parameter input array is not the right size'
                ' for the number of model levels.')

    elif isinstance(values, (int, float)):
        return DataArray(values * np.ones(numlevels, ),
                         dims=('mid_levels',),
                         attrs={'units': units})

    raise TypeError(
        'Cloud variable input must be a single value, `numpy.ndarray` or a '
        '`sympl.DataArray`')


# TODO: Make this a staticmethod of the ``Cloud`` class?
def get_waveband_data_array(values, units='dimensionless', numlevels=200,
                            sw=True):
    """Return a DataArray of values."""
    if isinstance(values, DataArray):
        return values

    if sw:
        dims_bands = 'num_shortwave_bands'
        numbands = 14
    else:
        dims_bands = 'num_longwave_bands'
        numbands = 16

    if isinstance(values, (int, float)):
        return DataArray(values * np.ones((numlevels, numbands)),
                         dims=('mid_levels', dims_bands),
                         attrs={'units': units})

    elif isinstance(values, np.ndarray):
        if values.shape == (numlevels,):
            return DataArray(
                np.repeat(values[:, np.newaxis], numbands, axis=1),
                dims=('mid_levels', dims_bands),
                attrs={'units': units},
            )
        elif values.shape == (numlevels, numbands):
            return DataArray(
                values,
                dims=('mid_levels', dims_bands),
                attrs={'units': units},
            )

    raise TypeError(
        'Cloud variable input must be a single value, `numpy.ndarray` or a '
        '`sympl.DataArray`')


def get_rectangular_profile(z, value, ztop, depth):
    p = np.zeros(z.shape)
    inrectangle = np.logical_and(z < ztop, z > ztop - depth)
    p[inrectangle] = value

    return p


class Cloud(metaclass=abc.ABCMeta):
    """Base class to define abstract methods for all cloud handlers.
    Default properties include a cloud area fraction equal to zero everywhere
    (ie no cloud)."""
    num_longwave_bands = 16
    num_shortwave_bands = 14

    def __init__(self, numlevels, cloud_fraction=0, mass_ice=0, mass_water=0,
                 ice_particle_size=20, droplet_radius=10,
                 lw_optical_thickness=0, sw_optical_thickness=0,
                 forward_scattering_fraction=0, asymmetry_parameter=0.85,
                 single_scattering_albedo=0.9):
        """Create a cloud.

        Parameters:
            numlevels (int): Number of atmospheric levels.
            cloud_fraction (float / ndarray / DataArray): cloud area fraction
            mass_ice (float / ndarray / DataArray): mass content of cloud ice
                [kg m-2]
            mass_water (float / ndarray / DataArray): mass content of cloud
                liquid water [kg m-2]
            ice_particle_size (float / ndarray / DataArray): cloud ice particle
                size [micrometers]
            droplet_radius (float / ndarray / DataArray): cloud water droplet
                radius [micrometers]
            lw_optical_thickness (float / DataArray): longwave optical
                thickness of the cloud
            sw_optical_thickness (float / DataArray): shortwave optical
                thickness of the cloud
            forward_scattering_fraction (float / DataArray): cloud forward
                scattering fraction (for the shortwave component of RRTMG)
                This is a scaling factor for the other shortwave parameters,
                if it is set to 0, no scaling is applied.
            asymmetry_parameter (float / DataArray): cloud asymmetry parameter
                (for the shortwave component of RRTMG)
            single_scattering_albedo (float / DataArray): single scattering
                albedo due to cloud (for the shortwave component of RRTMG)
        """
        self.numlevels = numlevels

        self.mass_content_of_cloud_liquid_water_in_atmosphere_layer = \
            get_p_data_array(mass_water, numlevels=self.numlevels)

        self.mass_content_of_cloud_ice_in_atmosphere_layer = get_p_data_array(
            mass_ice, numlevels=self.numlevels)

        self.cloud_area_fraction_in_atmosphere_layer = get_p_data_array(
            cloud_fraction, numlevels=self.numlevels, units='dimensionless')

        self.cloud_ice_particle_size = get_p_data_array(
            ice_particle_size, numlevels=self.numlevels, units='micrometers')

        self.cloud_water_droplet_radius = get_p_data_array(
            droplet_radius, numlevels=self.numlevels, units='micrometers')

        self.longwave_optical_thickness_due_to_cloud = get_waveband_data_array(
            lw_optical_thickness, numlevels=self.numlevels, sw=False)

        self.cloud_forward_scattering_fraction = get_waveband_data_array(
            forward_scattering_fraction, numlevels=self.numlevels)

        self.cloud_asymmetry_parameter = get_waveband_data_array(
            asymmetry_parameter,numlevels=self.numlevels)

        self.shortwave_optical_thickness_due_to_cloud = \
            get_waveband_data_array(sw_optical_thickness,
                                    numlevels=self.numlevels)

        self.single_scattering_albedo_due_to_cloud = get_waveband_data_array(
            single_scattering_albedo, numlevels=self.numlevels)

    @classmethod
    def from_atmosphere(cls, atmosphere, **kwargs):
        """Initialize a cloud component matching the given atmosphere.

        Parameters:
            atmosphere (``konrad.atmosphere.Atmosphere``):
                Atmosphere component.

        """
        return cls(numlevels=atmosphere['plev'].size, **kwargs)

    @abc.abstractmethod
    def update_cloud_profile(self, atmosphere, convection, **kwargs):
        """Return the cloud parameters for the radiation scheme.
        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere): atmosphere model
            convection (konrad.convection): convection scheme
        """


class ClearSky(Cloud):
    def update_cloud_profile(self, *args, **kwargs):
        return


class PhysicalCloud(Cloud):
    """Define a cloud based on physical properties.

    The physical properties are cloud ice and liquid mass (per model level)
    and particle size.  To be used with
    cloud_optical_properties='liquid_and_ice_clouds' in climt/RRTMG.
    """
    def __init__(self, numlevels, cloud_fraction, mass_water, mass_ice,
                 ice_particle_size, droplet_radius):
        """
        Parameters:
        z (ndarray): an array with the size of the model levels
        cloud_fraction (float / ndarray / DataArray): cloud area fraction
        mass_ice (float / ndarray / DataArray): mass content of cloud ice
            [kg m-2]
        mass_water (float / ndarray / DataArray): mass content of cloud
            liquid water [kg m-2]
        ice_particle_size (float / ndarray / DataArray): cloud ice particle
            size [micrometers]
        droplet_radius (float / ndarray / DataArray): cloud water droplet
            radius [micrometers]
        """
        super().__init__(
            numlevels=numlevels,
            cloud_fraction=cloud_fraction,
            mass_ice=mass_ice,
            mass_water=mass_water,
            ice_particle_size=ice_particle_size,
            droplet_radius=droplet_radius
        )

    def update_cloud_profile(self, *args, **kwargs):
        """Keep the cloud fixed with pressure. """
        return


class DirectInputCloud(Cloud):
    """ To be used with cloud_optical_properties='direct_input' in climt/RRTMG.
    """
    def __init__(self, numlevels, cloud_fraction, lw_optical_thickness,
                 sw_optical_thickness, forward_scattering_fraction=0,
                 asymmetry_parameter=0.85, single_scattering_albedo=0.9):

        super().__init__(
            numlevels=numlevels,
            cloud_fraction=cloud_fraction,
            lw_optical_thickness=lw_optical_thickness,
            sw_optical_thickness=sw_optical_thickness,
            forward_scattering_fraction=forward_scattering_fraction,
            asymmetry_parameter=asymmetry_parameter,
            single_scattering_albedo=single_scattering_albedo
        )

        self._norm_index = None
        self._interp_cldf = None
        self._interp_sw = None
        self._interp_lw = None

    def __add__(self, other):
        """Define the superposition of two clouds in a layer."""
        name_map = (
            ('cloud_fraction',
             'cloud_area_fraction_in_atmosphere_layer'),
            ('lw_optical_thickness',
             'longwave_optical_thickness_due_to_cloud'),
            ('sw_optical_thickness',
             'shortwave_optical_thickness_due_to_cloud'),
            ('forward_scattering_fraction',
             'cloud_forward_scattering_fraction'),
            ('asymmetry_parameter',
             'cloud_asymmetry_parameter'),
            ('single_scattering_albedo',
             'single_scattering_albedo_due_to_cloud'),
        )

        # The superposition of two clouds is implemented following a
        # "The winner takes it all"-approach:
        # For each cloud layer, the properties of the bigger cloud (in terms
        # of cloud fraction) is used.
        other_is_bigger = (
                other.cloud_area_fraction_in_atmosphere_layer
                > self.cloud_area_fraction_in_atmosphere_layer
        )

        kwargs = {}
        for kw_name, attr_name in name_map:
            arr = getattr(self, attr_name).values.copy()
            arr[other_is_bigger] = getattr(other, attr_name)[other_is_bigger]
            kwargs[kw_name] = arr

        summed_cloud = DirectInputCloud(numlevels=self.numlevels, **kwargs)

        return summed_cloud

    def interpolation_function(self, cloud_parameter):
        """ Calculate the interpolation function, to be used to maintain the
        cloud optical properties and keep the cloud attached to a normalisation
        level (self._norm_index). A separate interpolation function is required
        for each cloud parameter that needs to be interpolated.

        Parameters:
            cloud_parameter (DataArray): property to be interpolation
        Returns:
            scipy.interpolate.interpolate.interp1d
        """
        normed_levels = np.arange(0, self.numlevels) - self._norm_index

        interpolation_f = interp1d(
            normed_levels,
            cloud_parameter.values,
            fill_value='extrapolate',
            axis=0,
        )
        return interpolation_f

    def shift_property(self, cloud_parameter, interpolation_f, norm_new):
        """Shift the cloud area fraction according to a normalisation level.

        Parameters:
            cloud_parameter (DataArray): cloud property to be shifted
            interpolation_f (scipy.interpolate.interpolate.interp1d):
                interpolation object calculated by interpolation_function
            norm_new (int / float): normalisation index [model level]

        Returns:
            DataArray: shifted cloud property
        """
        levels = np.arange(0, self.numlevels)
        if norm_new is not np.nan:
            # Move the cloud to the new normalisation level, if there is one.
            cloud_parameter.values = interpolation_f(levels - norm_new)
        else:
            # Otherwise keep the cloud where it is.
            cloud_parameter.values = interpolation_f(levels - self._norm_index)

        return cloud_parameter

    def shift_cloud_profile(self, norm_new):
        if self._norm_index is None:
            self._norm_index = norm_new

            self._interp_cldf = self.interpolation_function(
                cloud_parameter=self.cloud_area_fraction_in_atmosphere_layer,
            )

            self._interp_sw = self.interpolation_function(
                cloud_parameter=self.shortwave_optical_thickness_due_to_cloud,
            )

            self._interp_lw = self.interpolation_function(
                cloud_parameter=self.longwave_optical_thickness_due_to_cloud,
            )

        self.cloud_area_fraction_in_atmosphere_layer = self.shift_property(
            cloud_parameter=self.cloud_area_fraction_in_atmosphere_layer,
            interpolation_f=self._interp_cldf,
            norm_new=norm_new,
        )

        self.shortwave_optical_thickness_due_to_cloud = self.shift_property(
            cloud_parameter=self.shortwave_optical_thickness_due_to_cloud,
            interpolation_f=self._interp_sw,
            norm_new=norm_new,
        )

        self.longwave_optical_thickness_due_to_cloud = self.shift_property(
            cloud_parameter=self.longwave_optical_thickness_due_to_cloud,
            interpolation_f=self._interp_lw,
            norm_new=norm_new,
        )

    def update_cloud_profile(self, atmosphere, convection, **kwargs):
        """Keep the cloud profile fixed with model level (pressure). """
        return


class HighCloud(DirectInputCloud):
    """Representation of a high-level cloud."""
    def update_cloud_profile(self, atmosphere, convection, **kwargs):
        """Keep the cloud attached to the convective top. """
        self.shift_cloud_profile(
            norm_new=convection.get('convective_top_index')[0],
        )


class MidLevelCloud(DirectInputCloud):
    """Representation of a mid-level cloud."""
    def update_cloud_profile(self, atmosphere, convection, **kwargs):
        """Keep the cloud attached to the freezing point. """
        self.shift_cloud_profile(
            norm_new=atmosphere.get_triple_point_index(),
        )


class LowCloud(DirectInputCloud):
    """Representation of a low-level cloud.
    Fixed at the top of the planetary boundary layer."""
    def __init__(self, numlevels, height_of_cloud, cloud_fraction,
                 lw_optical_thickness, sw_optical_thickness,
                 forward_scattering_fraction=0, asymmetry_parameter=0.85,
                 single_scattering_albedo=0.9):
        """
        Additional parameters to parent class:
            height_of_cloud (float/int): height at which the cloud is fixed [m]
        """
        super().__init__(
            numlevels=numlevels,
            cloud_fraction=cloud_fraction,
            lw_optical_thickness=lw_optical_thickness,
            sw_optical_thickness=sw_optical_thickness,
            forward_scattering_fraction=forward_scattering_fraction,
            asymmetry_parameter=asymmetry_parameter,
            single_scattering_albedo=single_scattering_albedo
        )

        self._z_of_cloud = height_of_cloud

    def update_cloud_profile(self, atmosphere, convection, **kwargs):
        """Keep the cloud at a fixed height. """
        z = atmosphere.get('z')[-1]
        index = int(np.argmax(z > self._z_of_cloud))  # model level above
        height_array = np.array([z[index-1], z[index]])
        index_array = np.array([index-1, index])  # model level bounds
        # Interpolate between two indices for model levels to be closer to that
        # corresponding to the height at which the cloud should be fixed.
        norm_new = interp1d(height_array, index_array)(self._z_of_cloud)

        self.shift_cloud_profile(
            norm_new=norm_new
        )


class CloudEnsemble(DirectInputCloud):
    """Wrapper to combine several clouds into a cloud ensemble.

    Warning: For now, overlapping clouds are handled very poorly!

    A cloud ensemble can consist of an arbitrary number of clouds.
    After its initialization it is handled like a normal `Cloud`:

    >>> cloud1 = HighCloud(...)
    >>> cloud2 = LowCloud(...)
    >>> cloud_ensemble = CloudEnsemble(cloud1, cloud2)
    >>> cloud_ensemble.cloud_area_fraction_in_atmosphere_layer

    """
    def __init__(self, *args):
        if not all([isinstance(a, DirectInputCloud) for a in args]):
            raise ValueError(
                'Only `DirectInputCloud`s can be combined in an ensemble.')
        else:
            self.clouds = args

        self._superposition = None
        self.superpose()

    def __getattr__(self, name):
        return getattr(self._superposition, name)

    def __getitem__(self, name):
        return getattr(self, name)

    def superpose(self):
        """Update the superposed cloud profile."""
        self._superposition = np.sum(self.clouds)

    def update_cloud_profile(self, *args, **kwargs):
        """Update every cloud in the cloud ensemble."""
        for cloud in self.clouds:
            cloud.update_cloud_profile(*args, **kwargs)

        self.superpose()



"""Coupling mechanisms between tropospheric and stratospheric water vapor."""
import abc

import numpy as np

from konrad.component import Component


class StratosphereCoupler(Component, metaclass=abc.ABCMeta):
    """Define the coupling of tropospheric and stratospheric water vapor."""
    @abc.abstractmethod
    def adjust_stratospheric_vmr(self, atmosphere):
        """Adjust stratospheric water vapor VMR values.

        Note:
            This method modifies the values in the atmosphere model!

        Parameters:
            atmosphere (``konrad.atmosphere.Atmosphere``):
                Atmosphere component.
            """
        return


class ColdPointCoupling(StratosphereCoupler):
    """Keep stratospheric VMR constant from the cold point on."""
    def adjust_stratospheric_vmr(self, atmosphere):
        cp_index = atmosphere.get_cold_point_index()
        atmosphere['H2O'][-1, cp_index:] = atmosphere['H2O'][-1, cp_index]


class FixedStratosphericVMR(StratosphereCoupler):
    def __init__(self, stratospheric_vmr=5e-6):
        """Keep stratospheric VMR fixed at a constant value.

        Parameters:
            stratospheric_vmr (float): Stratospheric water vapor amount [VMR].
        """
        self.stratospheric_vmr = stratospheric_vmr

    def adjust_stratospheric_vmr(self, atmosphere):
        cp_index = atmosphere.get_cold_point_index()
        atmosphere['H2O'][-1, cp_index:] = self.stratospheric_vmr


class MinimumStratosphericVMR(StratosphereCoupler):
    def __init__(self, minimum_vmr=5e-6):
        """Prevent stratospheric VMR from deceding a fixed minimum.

        Parameters:
            minimum_vmr (float): Minimum water vapor amount [VMR].
        """
        self.minimum_vmr = minimum_vmr

    def adjust_stratospheric_vmr(self, atmosphere):
        vmr = atmosphere['H2O'][-1, :]

        # If the VMR falls below the stratospheric background...
        if np.any(vmr < self.minimum_vmr):
            # ... set all values equal to the background from there.
            vmr[np.argmax(vmr < self.minimum_vmr):] = self.minimum_vmr
        else:
            # Otherwise find the smallest VMR and use the background value
            # from there on, this at least minimizes the discontinuity at
            # the transition point.
            vmr[np.argmin(vmr):] = self.minimum_vmr


class NoCoupling(StratosphereCoupler):
    """Do not adjust stratospheric water vapor.

    This coupler does not change the stratospheric water vapor at all. It
    may be used together with ``konrad.humidityy.FixedVMR()`` to perform
    simulations for a given set of VMR values (e.g. observations).

    Warning:
        This may lead to unrealistic atmospheric states when using models to
        describe the vertical relative humidity distribution!
    """
    def adjust_stratospheric_vmr(self, atmosphere):
        return



# -*- coding: utf-8 -*-
"""This module contains classes handling different convection schemes."""
import abc

import numpy as np
import typhon
from scipy.interpolate import interp1d

from konrad import constants
from konrad.component import Component
from konrad.surface import SurfaceFixedTemperature


__all__ = [
    'Convection',
    'NonConvective',
    'HardAdjustment',
    'RelaxedAdjustment',
]


def energy_difference(T_2, T_1, sst_2, sst_1, dp, eff_Cp_s):
    """
    Calculate the energy difference between two atmospheric profiles (2 - 1).

    Parameters:
        T_2: atmospheric temperature profile (2)
        T_1: atmospheric temperature profile (1)
        sst_2: surface temperature (2)
        sst_1: surface temperature (1)
        dp: pressure thicknesses of levels,
            must be the same for both atmospheric profiles
        eff_Cp_s: effective heat capacity of surface
    """
    Cp = constants.isobaric_mass_heat_capacity
    g = constants.g

    dT = T_2 - T_1  # convective temperature change of atmosphere
    dT_s = sst_2 - sst_1  # of surface

    termdiff = - np.sum(Cp/g * dT * dp) + eff_Cp_s * dT_s

    return termdiff


class Convection(Component, metaclass=abc.ABCMeta):
    """Base class to define abstract methods for convection schemes."""
    @abc.abstractmethod
    def stabilize(self, atmosphere, lapse, surface, timestep):
        """Stabilize the temperature profile by redistributing energy.

        Parameters:
              atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.
              lapse (ndarray): Temperature lapse rate [K/day].
              surface (konrad.surface): Surface model.
              timestep (float): Timestep width [day].
        """


class NonConvective(Convection):
    """Do not apply convection."""
    def stabilize(self, *args, **kwargs):
        pass


class HardAdjustment(Convection):
    """Instantaneous adjustment of temperature profiles"""
    def stabilize(self, atmosphere, lapse, surface, timestep):

        T_rad = atmosphere['T'][0, :]
        p = atmosphere['plev']

        # Find convectively adjusted temperature profile.
        T_new, T_s_new = self.convective_adjustment(
            p=p,
            phlev=atmosphere['phlev'],
            T_rad=T_rad,
            lapse=lapse,
            surface=surface,
            timestep=timestep,
        )
        # get convective top temperature and pressure
        self.calculate_convective_top(T_rad, T_new, p, timestep=timestep)
        # Update atmospheric temperatures as well as surface temperature.
        atmosphere['T'][0, :] = T_new
        surface['temperature'][0] = T_s_new

    def convective_adjustment(self, p, phlev, T_rad, lapse, surface,
                              timestep=0.1):
        """
        Find the energy-conserving temperature profile using a iterative
        procedure with test profiles. Update the atmospheric temperature
        profile to this one.

        Parameters:
            p (ndarray): pressure levels
            phlev (ndarray): half pressure levels
            T_rad (ndarray): old atmospheric temperature profile
            lapse (konrad.lapserate): lapse rate in K/km
            surface (konrad.surface):
                surface associated with old temperature profile
            timestep (float): only required for slow convection
        """
        # The threshold is scaled with the effective heat capacity of the
        # surface. Otherwise very thick surfaces may never reach the target.
        try:
            near_zero = float(surface.heat_capacity / 1e13)
        except KeyError:
            # heat_capacity is not defined for fixed temperature surfaces
            near_zero = 10**-8

        # Interpolate density and lapse rate on pressure half-levels.
        density1 = typhon.physics.density(p, T_rad)
        density = interp1d(p, density1, fill_value='extrapolate')(phlev[:-1])

        g = constants.earth_standard_gravity
        lp = -lapse / (g * density)

        # find energy difference if there is no change to surface temp due to
        # convective adjustment. in this case the new profile should be
        # associated with an increase in energy in the atmosphere.
        surfaceTpos = surface['temperature']
        T_con, diffpos = self.test_profile(T_rad, p, phlev, surface,
                                           surfaceTpos, lp,
                                           timestep=timestep)

        # this is the temperature profile required if we have a set-up with a
        # fixed surface temperature, then the energy does not matter.
        if isinstance(surface, SurfaceFixedTemperature):
            return T_con, surface['temperature']
        # for other cases, if we find a decrease or approx no change in energy,
        # the atmosphere is not being warmed by the convection,
        # as it is not unstable to convection, so no adjustment is applied
        if diffpos < near_zero:
            return T_con, surface['temperature']

        # if the atmosphere is unstable to convection, a fixed surface temp
        # produces an increase in energy (as convection warms the atmosphere).
        # this surface temperature is an upper bound to the energy-conserving
        # surface temperature.
        # taking the surface temperature as the coldest temperature in the
        # radiative profile gives us a lower bound.
        surfaceTneg = np.array([np.min(T_rad)])
        eff_Cp_s = surface.heat_capacity
        diffneg = eff_Cp_s * (surfaceTneg - surface['temperature'])
        # good guess for energy-conserving profile (unlikely!)
        if np.abs(diffneg) < near_zero:
            return T_con, surfaceTneg

        # NOTE (lkluft): Dirty workaround to always initialize `surfaceT`.
        # I encountered situations where the while-loop did not run a single
        # iteration and therefore the return-statement failed.
        surfaceT = (surfaceTneg + (surfaceTpos - surfaceTneg)
                    * (-diffneg) / (-diffneg + diffpos))

        # Now we have a upper and lower bound for the surface temperature of
        # the energy conserving profile. Iterate to get closer to the energy-
        # conserving temperature profile.
        counter = 0
        while diffpos >= near_zero and -diffneg >= near_zero:
            surfaceT = (surfaceTneg + (surfaceTpos - surfaceTneg)
                        * (-diffneg) / (-diffneg + diffpos))
            T_con, diff = self.test_profile(T_rad, p, phlev, surface, surfaceT,
                                            lp, timestep=timestep)
            if diff > 0:
                diffpos = diff
                surfaceTpos = surfaceT
            else:
                diffneg = diff
                surfaceTneg = surfaceT

            # to avoid getting stuck in a loop if something weird is going on
            counter += 1
            if counter == 100:
                raise ValueError(
                    "No energy conserving convective profile can be found"
                )

        # save new temperature profile
        return T_con, surfaceT

    def test_profile(self, T_rad, p, phlev, surface, surfaceT, lp,
                     timestep=0.1):
        """
        Assuming a particular surface temperature (surfaceT), create a new
        profile, following the specified lapse rate (lp) for the region where
        the convectively adjusted atmosphere is warmer than the radiative one.

        Parameters:
            T_rad (ndarray): old atmospheric temperature profile
            p (ndarray): pressure levels
            phlev (ndarray): half pressure levels
            surface (konrad.surface):
                surface associated with old temperature profile
            surfaceT (float): surface temperature of the new profile
            lp (ndarray): lapse rate in K/Pa
            timestep (float): not required in this case

        Returns:
            ndarray: new atmospheric temperature profile
            float: energy difference between the new profile and the old one
        """
        # dp, thicknesses of atmosphere layers, for energy calculation
        dp = np.diff(phlev)
        # for lapse rate integral
        dp_lapse = np.hstack((np.array([p[0] - phlev[0]]), np.diff(p)))
        T_con = surfaceT - np.cumsum(dp_lapse * lp)
        if np.any(T_con > T_rad):
            contop = np.max(np.where(T_con > T_rad))
            T_con[contop+1:] = T_rad[contop+1:]
        else:
            T_con = T_rad

        # If run with a fixed surface temperature, always return the
        # convective profile starting from the current surface temperature.
        if isinstance(surface, SurfaceFixedTemperature):
            return T_con, 0.

        eff_Cp_s = surface.heat_capacity

        diff = energy_difference(T_con, T_rad, surfaceT,
                                 surface['temperature'], dp, eff_Cp_s)

        return T_con, float(diff)

    def calculate_convective_top(self, T_rad, T_con, p, timestep=0.1, lim=0.2):
        """Find the pressure where the radiative heating has a certain value.

        Note:
            In the HardAdjustment case, for a contop temperature that is not
            dependent on the number of distribution of pressure levels, it is
            better to take a value of lim not equal or very close to zero.

        Parameters:
            T_rad (ndarray): radiative temperature profile [K]
            T_con (ndarray): convectively adjusted temperature profile [K]
            p (ndarray): model pressure levels [Pa]
            timestep (float): model timestep [days]
            lim (float): Threshold value [K/day].

        Returns:
            float: Pressure at height of convective top [Pa].
        """
        convective_heating = (T_con - T_rad) / timestep
        # Convective heating must be positive somewhere
        if np.any(convective_heating > lim):
            # NOTE: `np.argmax` returns the first occurrence of the maximum value.
            # In this example, the index of the first `True` value,
            # corresponding to the convective top, is returned.
            # The convective top corresponds to the first near zero / negative
            # convective heating rate above the region which is being heated
            # (the region which has a positive convective heating rate).
            positive_i = int(np.argmax(convective_heating > lim))
            contop_i = int(np.argmax(
                convective_heating[positive_i:] < lim)) + positive_i

            # Create auxiliary arrays storing the Qr, T and p values above and
            # below the threshold value. These arrays are used as input for the
            # interpolation in the next step.
            heat_array = np.array([convective_heating[contop_i-1],
                                   convective_heating[contop_i]])
            p_array = np.array([p[contop_i-1], p[contop_i]])
            T_array = np.array([T_con[contop_i-1], T_con[contop_i]])
            index_array = np.array([contop_i-1, contop_i])

            # Interpolate the pressure value to where the convective heating rate
            # equals `lim`.
            contop_index = interp1d(heat_array, index_array)(lim)
            contop_p = interp1d(heat_array, p_array)(lim)
            contop_T = interp1d(heat_array, T_array)(lim)

        else:
            convective_heating = np.zeros(p.shape)
            contop_index = np.nan
            contop_p = np.nan
            contop_T = np.nan

        self.create_variable('convective_heating_rate', convective_heating)
        self.create_variable('convective_top_plev', np.array([contop_p]))
        self.create_variable('convective_top_temperature', np.array([contop_T]))
        self.create_variable('convective_top_index', np.array([contop_index]))

        return

    def calculate_convective_top_height(self, z, lim=0.2):
        convective_heating = self.get('convective_heating_rate')[0]
        if not np.allclose(convective_heating, np.zeros(z.shape)):
            contop_i = int(np.argmax(convective_heating < lim))
            heat_array = np.array([convective_heating[contop_i - 1],
                                  convective_heating[contop_i]])
            z_array = np.array([z[contop_i - 1], z[contop_i]])
            contop_z = interp1d(heat_array, z_array, fill_value='extrapolate')(lim)
        else:
            contop_z = np.nan
        self.create_variable('convective_top_height', [contop_z])
        return

class RelaxedAdjustment(HardAdjustment):
    """Adjustment with relaxed convection in upper atmosphere.

    This convection scheme allows for a transition regime between a
    convectively driven troposphere and the radiatively balanced stratosphere.
    """
    def __init__(self, tau=None):
        """
        Parameters:
            tau (ndarray): Array of convective timescale values [days]
        """
        self.convective_tau = tau

    def get_convective_tau(self, p):
        """Return a convective timescale profile.

        Parameters:
            p (ndarray): Pressure levels [Pa].

        Returns:
            ndarray: Convective timescale profile [days].
        """
        if self.convective_tau is not None:
            return self.convective_tau

        tau0 = 1/24 # 1 hour
        tau = tau0*np.exp(p[0] / p)

        return tau

    def test_profile(self, T_rad, p, phlev, surface, surfaceT, lp,
                     timestep=0.1):
        """
        Assuming a particular surface temperature (surfaceT), create a new
        profile, using the convective timescale and specified lapse rate (lp).

        Parameters:
            T_rad (ndarray): old atmospheric temperature profile
            p (ndarray): pressure levels
            phlev (ndarray): half pressure levels
            surface (konrad.surface):
                surface associated with old temperature profile
            surfaceT (float): surface temperature of the new profile
            lp (ndarray): lapse rate in K/Pa
            timestep (float): not required in this case

        Returns:
            ndarray: new atmospheric temperature profile
            float: energy difference between the new profile and the old one
        """
        dp = np.diff(phlev)
        dp_lapse = np.hstack((np.array([p[0] - phlev[0]]), np.diff(p)))

        tau = self.get_convective_tau(p)

        tf = 1 - np.exp(-timestep / tau)
        T_con = T_rad * (1 - tf) + tf * (surfaceT - np.cumsum(dp_lapse * lp))

        # If run with a fixed surface temperature, always return the
        # convective profile starting from the current surface temperature.
        if isinstance(surface, SurfaceFixedTemperature):
            return T_con, 0.

        eff_Cp_s = surface.heat_capacity

        diff = energy_difference(T_con, T_rad, surfaceT,
                                 surface['temperature'], dp, eff_Cp_s)
        return T_con, float(diff)





# -*- coding: utf-8 -*-
"""This module contains classes for an upwelling induced cooling term."""
import abc

import numpy as np

from konrad import constants
from konrad.component import Component


class Upwelling(Component, metaclass=abc.ABCMeta):
    """Base class to define abstract methods for all upwelling handlers."""

    @abc.abstractmethod
    def cool(self, atmosphere, radheat, timestep):
        """ Cool the atmosphere according to an upwelling.
        Parameters:
              atmosphere (konrad.atmosphere.Atmosphere): Atmosphere model.
              radheat (ndarray): Radiative heatingrate [K/day].
              timestep (float): Timestep width [day].
        """


class NoUpwelling(Upwelling):
    """Do not apply a dynamical cooling."""
    def cool(self, *args, **kwargs):
        pass


class StratosphericUpwelling(Upwelling):
    """Apply a dynamical cooling, based on a specified upwelling velocity."""
    def __init__(self, w=0.2, lowest_level=None):
        """Create a upwelling handler.

        Parameters:
            w (float): Upwelling velocity in mm/s.
            lowest_level (int or None): The index of the lowest level to which
                the upwelling is applied. If none, uses the top of convection.
        """
        self.w = w * 86.4  # in m/day
        self.lowest_level = lowest_level

    def cool(self, atmosphere, convection, timestep):
        """Apply cooling above the convective top (level where the net
        radiative heating becomes small)."""

        Q = self.coolingrates(atmosphere)
        T = atmosphere['T'][0, :]

        if self.lowest_level is not None:
            contopi = self.lowest_level
        else:
            contopi = convection.get('convective_top_index')[0]
            if contopi is np.nan:
                return
        contopi = int(np.round(contopi))
        T_new = T[contopi:] + Q[contopi:] * timestep
        atmosphere['T'][0, contopi:] = T_new

    def coolingrates(self, atmosphere):
        """Get cooling rates associated with the upwelling velocity w."""
        dTdz = np.gradient(atmosphere['T'][0, :], atmosphere['z'][0, :])

        g = constants.g
        Cp = constants.Cp
        Q = -self.w * (dTdz + g/Cp)

        return Q



# -*- coding: utf-8 -*-
"""Physical constants.
"""
import scipy.constants as spc
import typhon.constants as tyc


# Phyiscal constants
Cp = isobaric_mass_heat_capacity = 1003.5  # J kg^-1 K^-1
g = earth_standard_gravity = spc.g  # m s^-2
stefan_boltzmann = 5.67e-8  # W m^-2 K^-4
heat_of_vaporization = Lv = 2501000  # J k^-1
specific_gas_constant_dry_air = Rd = tyc.gas_constant_dry_air  # J kg^-1 K^-1
specific_gas_constant_water_vapor = Rv = tyc.gas_constant_water_vapor  # J kg^-1 K^-1
gas_constant_ratio = epsilon = Rd / Rv  # 1
density_sea_water = 1025  # kg m^-3
specific_heat_capacity_sea_water = 4185.5  # J kg^-1 K^-1
molar_gas_constant_dry_air = molar_Rd = spc.gas_constant  # J mol^-1 K^-1
avogadro = tyc.avogadro  # molecules per mole
triple_point_water = 273.16  # K

# Variable descriptions
variable_description = {
    # Atmospheric variables
    'plev': {
        'units': 'Pa',
        'standard_name': 'air_pressure',
        },
    'phlev': {
        'units': 'Pa',
        'standard_name': 'air_pressure_at_halflevel',
    },
    'time': {
        'standard_name': 'time',
        'units': 'hours since 0001-01-01 00:00:00.0',
        'calender': 'gregorian',
        },
    'T': {
        'units': 'K',
        'standard_name': 'air_temperature',
        'arts_name': 'T',
        'dims': ('time', 'plev'),
        },
    'z': {
        'units': 'm',
        'standard_name': 'geopotential_height',
        'description': 'Geopotential height calculated from atmospheric state',
        'arts_name': 'z',
        'dims': ('time', 'plev'),
        },
    'lapse': {
        'units': 'K / m',
        'standard_name': 'air_temperature_lapse_rate',
    },
    'H2O': {
        'units': '1',
        'standard_name': 'humidity_mixing_ratio',
        'arts_name': 'abs_species-H2O',
        'dims': ('time', 'plev'),
        },
    'N2O': {
        'units': '1',
        'standard_name': 'nitrogene_mixing_ratio',
        'arts_name': 'abs_species-N2O',
        'dims': ('plev',),
        'default_vmr': 306e-9,
        },
    'O3': {
        'units': '1',
        'standard_name': 'ozone_mixing_ratio',
        'arts_name': 'abs_species-O3',
        'dims': ('time', 'plev'),
        },
    'O2': {
        'units': '1',
        'standard_name': 'oxygen_mixing_ratio',
        'arts_name': 'abs_species-O2',
        'dims': ('plev',),
        'default_vmr': 0.21,
    },
    'CO2': {
        'units': '1',
        'standard_name': 'carbon_dioxide_mixing_ratio',
        'arts_name': 'abs_species-CO2',
        'dims': ('plev',),
        'default_vmr': 348e-6,
        },
    'CO': {
        'units': '1',
        'standard_name': 'carbon_monoxide_mixing_ratio',
        'arts_name': 'abs_species-CO',
        'dims': ('plev',),
        'default_vmr': 0.,
        },
    'CH4': {
        'units': '1',
        'standard_name': 'methane_mixing_ratio',
        'arts_name': 'abs_species-CH4',
        'dims': ('plev',),
        'default_vmr': 1650e-9,
        },
    'CFC11': {
        'units': '1',
        'standard_name': 'cfc11_mixing_ratio',
        'arts_name': 'abs_species-CFC11',
        'dims': ('plev',),
        'default_vmr': 0.,
    },
    'CFC12': {
        'units': '1',
        'standard_name': 'cfc12_mixing_ratio',
        'arts_name': 'abs_species-CFC12',
        'dims': ('plev',),
        'default_vmr': 0.,
    },
    'CFC22': {
        'units': '1',
        'standard_name': 'cfc22_mixing_ratio',
        'arts_name': 'abs_species-CFC22',
        'dims': ('plev',),
        'default_vmr': 0.,
    },
    'CCl4': {
        'units': '1',
        'standard_name': 'carbon_tetrachloride_mixing_ratio',
        'arts_name': 'abs_species-CCl4',
        'dims': ('plev',),
        'default_vmr': 0.,
    },
    'diabatic_convergence_max_plev': {
        'units': 'Pa',
        'standard_name': 'diabatic_convergence_max_plev',
        'description': 'Pressure level of maximum diabatic convergence',
        'dims': ('time',),
    },
    'diabatic_convergence_max_index': {
        'units': '1',
        'standard_name': 'diabatic_convergence_max_index',
        'description': 'Level index of maximum diabatic convergence',
        'dims': ('time',),
    },
    # Convective quantities
    'convective_heating_rate': {
        'units': 'K / day',
        'standard_name': 'convective_heating_rate',
        'dims': ('time', 'plev'),
    },
    'convective_top_plev': {
        'units': 'Pa',
        'standard_name': 'convective_top_plev',
        'description': 'Pressure level of the top of convection',
        'dims': ('time',),
    },
    'convective_top_temperature': {
        'units': 'K',
        'standard_name': 'convective_top_temperature',
        'description': 'Temperature at the top of convection',
        'dims': ('time',),
    },
    'convective_top_index': {
        'units': 'model_level_index',
        'standard_name': 'convective_top_index',
        'description': 'Model level index at the top of convection '
                       '(interpolated, between levels)',
        'dims': ('time',),
    },
    'convective_top_height': {
        'units': 'm',
        'standard_name': 'convective_top_height',
        'description': 'Altitude of the top of convection',
        'dims': ('time',),
    },
    # Radiative quantities
    'lw_htngrt': {
        'units': 'K / day',
        'standard_name': 'tendency_of_air_temperature_due_to_longwave_heating',
        },
    'lw_htngrt_clr': {
        'units': 'K / day',
        'standard_name': ('tendency_of_air_temperature_'
                          'due_to_longwave_heating_assuming_clear_sky'
                          ),
        },
    'lw_flxu': {
        'units': 'W / m**2',
        'standard_name': 'upwelling_longwave_flux_in_air',
        },
    'lw_flxd': {
        'units': 'W / m**2',
        'standard_name': 'downwelling_longwave_flux_in_air',
        },
    'lw_flxu_clr': {
        'units': 'W / m**2',
        'standard_name': 'upwelling_longwave_flux_in_air_assuming_clear_sky',
        },
    'lw_flxd_clr': {
        'units': 'W / m**2',
        'standard_name': 'downwelling_longwave_flux_in_air_assuming_clear_sky',
        },
    'sw_htngrt': {
        'units': 'K / day',
        'standard_name':
            'tendency_of_air_temperature_due_to_shortwave_heating',
        },
    'sw_htngrt_clr': {
        'units': 'K / day',
        'standard_name': ('tendency_of_air_temperature_'
                          'due_to_shortwave_heating_assuming_clear_sky'
                          ),
        },
    'sw_flxu': {
        'units': 'W / m**2',
        'standard_name': 'upwelling_shortwave_flux_in_air',
        },
    'sw_flxd': {
        'units': 'W / m**2',
        'standard_name': 'downwelling_shortwave_flux_in_air',
        },
    'sw_flxu_clr': {
        'units': 'W / m**2',
        'standard_name': 'upwelling_shortwave_flux_in_air_assuming_clear_sky',
        },
    'sw_flxd_clr': {
        'units': 'W / m**2',
        'standard_name':
            'downwelling_shortwave_flux_in_air_assuming_clear_sky',
        },
    'net_htngrt': {
        'units': 'K / day',
        'standard_name':
            'tendency_of_air_temperature_due_to_radiative_heating',
        },
    'toa': {
        'units': 'W / m**2',
        'standard_name':
            'radiation_budget_at_top_of_the_atmosphere',
    },
    'deltaT': {
        'units': 'K / day',
        'standard_name': 'tendency_of_air_temperature',
        },
    # Surface parameters
    'albedo': {
        'units': '1',
        'standard_name': 'surface_albedo',
    },
    'pressure': {
        'units': 'Pa',
        'standard_name': 'surface_pressure',
    },
    'temperature': {
        'units': 'K',
        'standard_name': 'surface_temperature',
    },
    'c_p': {
        'units': 'J / kg / K',
        'standard_name': 'specific_heat_capacity_sea_water',
    },
    'rho': {
        'units': 'kg / m**3',
        'standard_name': 'surface_density',
    },
    'depth': {
        'units': 'm',
        'standard_name': 'surface_depth',
    },
    'heat_capacity': {
        'units': 'J / K',
        'standard_name': 'surface_heat_capacity',
    },
    # Ozone quantities
    'ozone_source': {
        'units': 'ppv / day',
        'dims': ('time', 'plev'),
    },
    'ozone_sink_ox': {
        'units': 'ppv / day',
        'dims': ('time', 'plev'),
    },
    'ozone_sink_nox': {
        'units': 'ppv / day',
        'dims': ('time', 'plev'),
    },
    'ozone_transport': {
        'units': 'ppv / day',
        'dims': ('time', 'plev'),
    },
    'ozone_sink_hox': {
        'units': 'ppv / day',
        'dims': ('time', 'plev'),
    }
}



import logging

import typhon
import netCDF4
import numpy as np
from scipy.interpolate import interp1d

from konrad import constants
from konrad import utils
from konrad.component import Component

__all__ = [
    'Atmosphere',
]

logger = logging.getLogger(__name__)


class Atmosphere(Component):
    """Atmosphere component.

    Attributes:
        atmosphere_variables (list[str]): Atmospheric variables defined by the
            ``Atmosphere`` component.
        pmin (float): Minimum pressure used as threshold between upper and
            lower atmosphere [Pa]. Methods like ``get_cold_point_index`` or
            ``get_triple_point_index`` are looking for levels with higher
            pressure (closer to the surface) only.
    """
    atmosphere_variables = [
        'T',
        'H2O',
        'N2O',
        'O3',
        'O2',
        'CO2',
        'CO',
        'CH4',
        'CFC11',
        'CFC12',
        'CFC22',
        'CCl4',
    ]
    pmin = 10e2

    def __init__(self, plev):
        """Initialise atmosphere component.

        Parameters:
            plev (``np.ndarray``): Atmospheric pressure (surface to top) [Pa].
        """
        super().__init__()

        self.coords = {
            'time': np.array([]),  # time dimension
            'plev': plev,  # pressure at full-levels
            # TODO: Should `phlev` be input?
            'phlev': utils.phlev_from_plev(plev)  # pressure at halflevels
        }

        for varname in self.atmosphere_variables:
            self.create_variable(varname, np.zeros_like(plev))

        # TODO: Combine with ``tracegases_rcemip``?
        self.create_variable(
            name='T',
            data=utils.standard_atmosphere(plev, coordinates='pressure'),
        )
        self.update_height()

        self.tracegases_rcemip()

    @classmethod
    def from_atm_fields_compact(cls, atm_fields_compact):
        """Convert an ARTS atm_fields_compact [0] into an atmosphere.

        [0] http://arts.mi.uni-hamburg.de/docserver-trunk/variables/atm_fields_compact

        Parameters:
            atm_fields_compact (typhon.arts.types.GriddedField4):
                Compact set of atmospheric fields.
        """

        def _extract_profile(atmfield, species):
            try:
                arts_key = constants.variable_description[species]['arts_name']
            except KeyError:
                logger.warning(f'No variabel description for "{species}".')
            else:
                return atmfield.get(arts_key, keep_dims=False)

        datadict = {var: _extract_profile(atm_fields_compact, var)
                    for var in cls.atmosphere_variables}
        datadict['plev'] = atm_fields_compact.grids[1]

        return cls.from_dict(datadict)

    @classmethod
    def from_xml(cls, xmlfile):
        """Read atmosphere from XML file containing an ARTS atm_fields_compact.

        Parameters:
            xmlfile (str): Path to XML file.
        """
        # Read the content of given XML file.
        griddedfield = typhon.arts.xml.load(xmlfile)

        # Check if the XML file contains an atm_fields_compact (GriddedField4).
        arts_type = typhon.arts.utils.get_arts_typename(griddedfield)
        if arts_type != 'GriddedField4':
            raise TypeError(
                'XML file contains "{}". Expected "GriddedField4".'.format(
                    arts_type)
            )

        return cls.from_atm_fields_compact(griddedfield, **kwargs)

    @classmethod
    def from_dict(cls, dictionary):
        """Create an atmosphere model from dictionary values.

        Parameters:
            dictionary (dict): Dictionary containing ndarrays.
        """
        # TODO: Currently working for good-natured dictionaries.
        #  Consider a more flexible user interface.

        # Create a Dataset with time and pressure dimension.
        plev = dictionary['plev']
        d = cls(plev=plev)

        for var in cls.atmosphere_variables:
            val = dictionary.get(var)
            if val is not None:
                # Prevent variables, that are not stored in the netCDF file,
                # to be overwritten with ``None``.
                d.create_variable(var, val)

        # Calculate the geopotential height.
        d.update_height()

        return d

    @classmethod
    def from_netcdf(cls, ncfile, timestep=-1):
        """Create an atmosphere model from a netCDF file.

        Parameters:
            ncfile (str): Path to netCDF file.
            timestep (int): Timestep to read (default is last timestep).
        """

        def _return_profile(ds, var, ts):
            return (ds[var][ts, :] if 'time' in ds[var].dimensions
                    else ds[var][:])

        with netCDF4.Dataset(ncfile) as root:
            if 'atmosphere' in root.groups:
                dataset = root['atmosphere']
            else:
                dataset = root

            datadict = {var: np.array(_return_profile(dataset, var, timestep))
                        for var in cls.atmosphere_variables
                        if var in dataset.variables
                        }
            datadict['plev'] = np.array(root['plev'][:])

        return cls.from_dict(datadict)

    def to_atm_fields_compact(self):
        """Convert an atmosphere into an ARTS atm_fields_compact."""
        # Store all atmosphere variables including geopotential height.
        variables = self.atmosphere_variables + ['z']

        # Get ARTS variable name from variable description.
        species = [constants.variable_description[var].get('arts_name')
                   for var in variables]

        # Create a GriddedField4.
        atmfield = typhon.arts.types.GriddedField4()

        # Set grids and their names.
        atmfield.gridnames = ['Species', 'Pressure', 'Longitude', 'Latitude']
        atmfield.grids = [
            species, self['plev'], np.array([]), np.array([])
        ]

        # The profiles have to be passed in "stacked" form, as an ndarray of
        # dimensions [species, pressure, lat, lon].
        atmfield.data = np.vstack(
            [self[var].reshape(1, self['plev'].size, 1, 1)
             for var in variables]
        )
        atmfield.dataname = 'Data'

        # Perform a consistency check of the passed grids and data tensor.
        atmfield.check_dimension()

        return atmfield

    def hash_attributes(self):
        """Create hash based on some basic characteristics"""
        return hash((
            self['plev'].min(),  # Pressure at top of the atmosphere
            self['plev'].max(),  # Surface pressure
            self['plev'].size,  # Number of pressure layers
            np.round(self['CO2'][0] / 1e-6),  # CO2 ppmv
            np.round(self['T'][-1, 0], 3),  # Surface temperature
        ))

    def refine_plev(self, pgrid, **kwargs):
        """Refine the pressure grid of an atmosphere object.

        Note:
              This method returns a **new** object,
              the original object is maintained!

        Parameters:
              pgrid (ndarray): New pressure grid [Pa].
            **kwargs: Additional keyword arguments are collected
                and passed to :func:`scipy.interpolate.interp1d`

        Returns:
              Atmosphere: A **new** atmosphere object.
        """
        # Initialize an empty directory to fill it with interpolated data.
        # The dictionary is later used to create a new object using the
        # Atmosphere.from_dict() classmethod. This allows to circumvent the
        # fixed dimension size in xarray.DataArrays.
        datadict = dict()

        datadict['plev'] = pgrid  # Store new pressure grid.

        # Loop over all atmospheric variables...
        for variable in self.atmosphere_variables:
            # and create an interpolation function using the original data.
            f = interp1d(self['plev'], self[variable],
                         axis=-1, fill_value='extrapolate', **kwargs)

            # Store the interpolated new data in the data directory.
            # dims = self.default_dimensions[variable]
            # datadict[variable] = DataArray(f(pgrid), dims=dims)
            datadict[variable] = f(pgrid).ravel()

        # Create a new atmosphere object from the filled data directory.
        # This method also calculates the new phlev coordinates.
        new_atmosphere = type(self).from_dict(datadict)

        # Keep attributes of original atmosphere object.
        # This is **extremely** important because references to e.g. the
        # convection scheme or the humidity handling are stored as attributes!
        new_atmosphere.attrs.update({**self.attrs})

        # Calculate the geopotential height.
        new_atmosphere.update_height()

        return new_atmosphere

    def calculate_height(self):
        """Calculate the geopotential height."""
        g = constants.earth_standard_gravity

        plev = self['plev']  # Air pressure at full-levels.
        phlev = self['phlev']  # Air pressure at half-levels.
        T = self['T']  # Air temperature at full-levels.

        rho = typhon.physics.density(plev, T)
        dp = np.hstack((np.array([plev[0] - phlev[0]]), np.diff(plev)))

        # Use the hydrostatic equation to calculate geopotential height from
        # given pressure, density and gravity.
        return np.cumsum(-dp / (rho * g))

    def update_height(self):
        """Update the value for height."""
        z = self.calculate_height()
        # If height is already in Dataset, update its values.
        if 'z' in self.data_vars:
            self.set('z', z)
        # Otherwise create the DataArray.
        else:
            self.create_variable('z', z)

    def get_cold_point_index(self):
        """Return the model level index at the cold point.

        Returns:
            int: Model level index at the cold point.
        """
        plev = self['plev'][:]
        T = self['T'][-1, :]

        return np.argmin(T[plev > self.pmin])

    def get_cold_point_plev(self):
        """Return the cold point pressure.

        Returns:
            float: Pressure at the cold point [Pa].
        """
        return self['plev'][self.get_cold_point_index()]

    def get_triple_point_index(self):
        """Return the model level index at the triple point.

        The triple point is taken at the temperature closest to 0 C.

        Returns:
            int: Model level index at the triple point.
        """
        plev = self['plev']
        T = self['T'][0, :]

        return np.argmin(np.abs(T[np.where(plev > self.pmin)] - 273.15))

    def get_triple_point_plev(self):
        """
        Return the pressure at the triple point.

        The triple point is taken at the temperature closest to 0 C.

        Returns:
            float: Pressure at the triple point [Pa].
        """
        return self['plev'][self.get_triple_point_index()]

    def get_lapse_rates(self):
        """Calculate the temperature lapse rate at each level."""
        return np.gradient(self['T'][0, :], self['z'][0, :])

    def get_potential_temperature(self, p0=1000e2):
        r"""Calculate the potential temperature.

        .. math::
            \theta = T \cdot \left(\frac{p_0}{P}\right)^\frac{2}{7}

        Parameters:
              p0 (float): Pressure at reference level [Pa].

        Returns:
              ndarray: Potential temperature [K].
        """
        # Get view on temperature and pressure arrays.
        T = self['T'][0, :]
        p = self['plev']

        # Calculate the potential temperature.
        return T * (p0 / p) ** (2 / 7)

    def get_static_stability(self):
        r"""Calculate the static stability.

        .. math::
            \sigma = - \frac{T}{\Theta} \frac{\partial\Theta}{\partial p}

        Returns:
              ndarray: Static stability [K/Pa].
        """
        # Get view on temperature and pressure arrays.
        t = self['T'][0, :]
        p = self['plev']

        # Calculate potential temperature and its vertical derivative.
        theta = self.get_potential_temperature()
        dtheta = np.gradient(theta, p)

        return -(t / theta) * dtheta

    def get_diabatic_subsidence(self, radiative_cooling):
        """Calculate the diabatic subsidence.

        Parameters:
              radiative_cooling (ndarray): Radiative cooling rates.
                Positive values for heating, negative values for cooling!

        Returns:
            ndarray: Diabatic subsidence [Pa/day].
        """
        sigma = self.get_static_stability()

        return -radiative_cooling / sigma

    def get_subsidence_convergence_max(self, radiative_cooling):
        """Return index of maximum subsidence convergence.

        Parameters:
            radiative_cooling (ndarray): Radiative cooling rates.
                Positive values for heating, negative values for cooling!
            pmin (float): Lower pressure threshold. The cold point has to
                be below (higher pressure, lower height) that value.

        Returns:
              float: Pressure of maxixum subsidence divergence [Pa].
        """
        plev = self['plev']
        omega = self.get_diabatic_subsidence(radiative_cooling)
        domega = np.gradient(omega, plev)

        max_index = np.argmax(domega[plev > self.pmin])
        max_plev = plev[max_index]

        self.create_variable('diabatic_convergence_max_index', [max_index])
        self.create_variable('diabatic_convergence_max_plev', [max_plev])

        return max_plev

    def tracegases_rcemip(self):
        """Set trace gas concentrations according to the RCE-MIP configuration.

        The volume mixing ratios are following the values for the
        RCE-MIP (Wing et al. 2017) and constant throughout the atmosphere.
        """
        self.update_height()

        concentrations = {
            'H2O': utils.humidity_profile_rcemip(self.get('z')),
            'CO2': 348e-6,
            'CH4': 1650e-9,
            'N2O': 306e-9,
            'CO': 0,
            'O3': utils.ozone_profile_rcemip(self.get('plev')),
            'CFC11': 0,
            'CFC12': 0,
            'CFC22': 0,
            'CCl4': 0,
        }

        for gas, vmr in concentrations.items():
            self.set(gas, vmr)



# -*- coding: utf-8 -*-
import sys
from os.path import (dirname, join)
from distutils.core import setup
from setuptools import find_packages


__version__ = open(join(dirname(__file__), 'konrad', 'VERSION')).read().strip()

if not sys.version_info >= (3, 5, 1):
    sys.exit('Only support Python version >=3.5.1\n'
             'Found version is {}'.format(sys.version))

setup(
    name='konrad',
    author='The konrad developers',
    version=__version__,
    url='https://github.com/atmtools/konrad',
    download_url='https://github.com/atmtools/konrad/tarball/v' + __version__,
    packages=find_packages(),
    license='MIT',
    description='Implementation of a radiative-convective equilibrium model.',
    classifiers=[
        # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=[
        'matplotlib>=2.0.0',
        'netcdf4>=1.2.7',
        'numpy>=1.12.0',
        'scipy>=0.19.0',
        'typhon>=0.7.0',
        'xarray>=0.9.1',
        'climt==0.16.0',
        'sympl>=0.4.0',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    extras_require={'tests': ['pytest']},
)



from os.path import (dirname, join)

import numpy as np
import pytest

from konrad import (atmosphere, utils)


@pytest.fixture
def atmosphere_obj():
    plev, _ = utils.get_pressure_grids(1000e2, 1, 50)

    return atmosphere.Atmosphere(plev=plev)


class TestAtmosphere:
    ref_dir = join(dirname(__file__), "reference_data", "")

    def test_init(self):
        """Test basic initialization of the atmosphere component."""
        plev = np.array([1000e2, 750e2, 500e2, 100e2, 10e2, 1e2])

        atmosphere.Atmosphere(plev=plev)

    def test_cold_point_index(self, atmosphere_obj):
        """Test retrieval of the cold point index."""
        assert atmosphere_obj.get_cold_point_index() == 11

    def test_triple_point_index(self, atmosphere_obj):
        """Test retrieval of the triple point index."""
        assert atmosphere_obj.get_triple_point_index() == 2

    def test_cold_point_plev(self, atmosphere_obj):
        """Test retrieval of the cold point pressure."""
        assert np.isclose(atmosphere_obj.get_cold_point_plev(), 20023.067)

    def test_triple_point_plev(self, atmosphere_obj):
        """Test retrieval of the triple point plev."""
        assert np.isclose(atmosphere_obj.get_triple_point_plev(), 74287.948)

    def test_from_netcdf(self):
        """Test initialisation from netCDF file."""
        ncfile = join(self.ref_dir, 'reference.nc')
        atmosphere.Atmosphere.from_netcdf(ncfile)

    def test_from_netcdf_invalid_input(self):
        """Check exception when passing unused keywords."""
        # Old versions of the ``Atmosphere`` API allowed to pass unused keyword
        # arguments to the ``from_*`` classmethods. This lead to unexpected
        # behaviour during initialisation.
        with pytest.raises(TypeError):
            atmosphere.Atmosphere.from_netcdf('dummy.nc', surface=None)



"""Different models to describe the vertical relative humidity distribution."""
import abc

import numpy as np
from scipy.stats import norm
from scipy.interpolate import interp1d

from konrad.component import Component


class RelativeHumidityModel(Component, metaclass=abc.ABCMeta):
    def __call__(self, atmosphere, **kwargs):
        """Return the vertical distirbution of relative humidity.

        Parameters:
            atmosphere (konrad.atmosphere.Atmosphere: Atmosphere component.
            **kwargs: Arbitrary number of additional arguments,
                depending on the actual implementation.

        Returns:
            ndarray: Relative humidity profile.
        """
        ...


class HeightConstant(RelativeHumidityModel):
    """Constant relative humidity profile throughout the whole troposphere."""
    def __init__(self, rh_surface=0.62):
        self.rh_surface = rh_surface
        self._rh_cache = None

    def __call__(self, atmosphere, **kwargs):
        if self._rh_cache is None:
            p = atmosphere['plev']
            self._rh_cache = self.rh_surface * np.ones_like(p)

        return self._rh_cache


class VerticallyUniform(RelativeHumidityModel):
    def __init__(self, rh_surface=0.5, rh_tropopause=0.3):
        self.rh_surface = rh_surface
        self.rh_tropopause = rh_tropopause
        self.convective_top = 300e2
        self.cold_point = 100e2

    def __call__(self, atmosphere, convection, **kwargs):
        p = atmosphere['plev']
        self.convective_top = convection.get('convective_top_plev')[0]
        self.cold_point = atmosphere.get_cold_point_plev()

        rh = (
            (self.rh_tropopause - self.rh_surface)
            / (self.cold_point - self.convective_top)
            * (p - self.convective_top) + self.rh_surface
        )
        rh[p > self.convective_top] = self.rh_surface

        return rh


class ConstantFreezingLevel(RelativeHumidityModel):
    """Constant rel. humidity up to the freezing level and then decreasing."""
    def __init__(self, rh_surface=0.77):
        self.rh_surface = rh_surface

    def __call__(self, atmosphere, **kwargs):
        plev = atmosphere['plev']
        rh_profile = self.rh_surface * np.ones_like(plev)

        fl = atmosphere.get_triple_point_index()
        rh_profile[fl:] = (
            self.rh_surface * (plev[fl:] / plev[fl])**1.3
        )

        return rh_profile


class FixedUTH(RelativeHumidityModel):
    """Idealised model of a fixed C-shaped relative humidity distribution."""
    def __init__(self, rh_surface=0.77, uth=0.75, uth_plev=170e2,
                 uth_offset=0):
        """Couple the upper-tropospheric humidity peak to the convective top.

        Parameters:
            rh_surface (float): Relative humidity at first
                pressure level (surface).
            uth (float): Relative humidity at the upper-tropospheric peak.
            uth_plev (float): Pressure level of second humidity maximum [Pa].
            uth_offset (float): Offset between UTH peak and convective top.
        """
        self.rh_surface = rh_surface
        self.uth = uth
        self.uth_plev = uth_plev
        self.uth_offset = uth_offset

        self._rh_base_profile = None

    def get_relative_humidity_profile(self, atmosphere):
        p = atmosphere['plev']

        # Use Manabe (1967) relative humidity model as base/background.
        if self._rh_base_profile is None:
            manabe_model = Manabe67(rh_surface=self.rh_surface)
            self._rh_base_profile = manabe_model(atmosphere)

        # Add skew-normal distribution.
        uth = self.uth * norm.pdf(
            x=np.log(p),
            loc=np.log(self.uth_plev + self.uth_offset),
            scale=0.4,
        )

        return np.maximum(self._rh_base_profile, uth)

    def __call__(self, atmosphere, **kwargs):
        return self.get_relative_humidity_profile(atmosphere)


class CoupledUTH(FixedUTH):
    """Idealised model of a coupled C-shaped relative humidity distribution.

    This relative humidity works in the same way as ``FixedUTH`` but the
    ``uth_plev`` is updated automatically depending on the convective top.
    """
    def __call__(self, atmosphere, convection, **kwargs):
        self.uth_plev = convection.get('convective_top_plev')[0]

        return self.get_relative_humidity_profile(atmosphere)


class CshapeConstant(RelativeHumidityModel):
    """Idealized model of a C-shaped RH profile using a quadratic equation."""
    def __init__(self, uth_plev=200e2, rh_min=0.3, uth=0.8):
        self.uth_plev = uth_plev
        self.rh_min = rh_min
        self.uth = uth
        self.rh_surface = uth

    def __call__(self, atmosphere, convection, **kwargs):
        self.uth_plev = convection.get('convective_top_plev')[0]

        x = np.log10(atmosphere['plev'])
        xmin = np.log10(self.uth_plev)
        xmax = x[0]

        a = (self.uth - self.rh_min) * 4 / (xmin - xmax)**2
        b = (xmin + xmax) / 2
        c = self.rh_min

        return np.clip(a * (x - b)**2 + c, a_min=0, a_max=1)


class CshapeDecrease(RelativeHumidityModel):
    """Idealized model of a C-shaped RH profile using a quadratic equation."""
    def __init__(self, uth_plev=200e2, rh_min=0.3, uth=0.8):
        self.uth_plev = uth_plev
        self.rh_min = rh_min
        self.uth = uth
        self.rh_surface = uth

    def __call__(self, atmosphere, convection, **kwargs):
        self.uth_plev = convection.get('convective_top_plev')[0]

        x = np.log10(atmosphere['plev'])
        xmin = np.log10(self.uth_plev)
        xmax = x[0]

        a = (self.uth - self.rh_min) * 4 / (xmin - xmax)**2
        b = (xmin + xmax) / 2
        c = self.rh_min

        rh = np.clip(a * (x - b)**2 + c, a_min=0, a_max=1)

        rh[x < xmin] *= (10**x / 10**xmin)[x < xmin]

        return rh


class Manabe67(RelativeHumidityModel):
    """Relative humidity model following Manabe and Wetherald (1967)."""
    def __init__(self, rh_surface=0.77):
        """Initialize a humidity model.

        Parameters:
            rh_surface (float): Relative humidity at the surface.
        """
        self.rh_surface = rh_surface

    def __call__(self, atmosphere, **kwargs):
        p = atmosphere['plev']

        return self.rh_surface * (p / p[0] - 0.02) / (1 - 0.02)


class Cess76(RelativeHumidityModel):
    """Relative humidity model following Cess (1976).

    The relative humidity profile depends on the surface temperature.
    This results in moister atmospheres at warmer temperatures.
    """
    def __init__(self, rh_surface=0.8, T_surface=288):
        """Initialize a humidity model.

        Parameters:
            rh_surface (float): Relative humidity at the surface.
            T_surface (float): Surface temperature [K].
        """
        self.rh_surface = rh_surface
        self.T_surface = T_surface

    @property
    def omega(self):
        """Temperature dependent scaling factor for the RH profile."""
        return 1.0 - 0.03 * (self.T_surface - 288)

    def __call__(self, atmosphere, surface, **kwargs):
        p = atmosphere['plev']
        self.T_surface = surface['temperature'][-1]

        return self.rh_surface * ((p / p[0] - 0.02) / (1 - 0.02))**self.omega


class Romps14(RelativeHumidityModel):
    """Return relative humidity according to an invariant RH-T relation."""
    def __call__(self, atmosphere, **kwargs):
        if self._rh_func is None:
            self._rh_func = interp1d(
                # Values read from Fig. 6 in Romps (2014).
                x=np.array([300, 240, 200, 190, 188, 186]),
                y=np.array([0.8, 0.6, 0.7, 1.0, 0.5, 0.1]),
                kind='linear',
                fill_value='extrapolate',
            )

        return self._rh_func(atmosphere['T'][-1, :])
